/***************************************************************************/
/* 		This code is part of Nscache - viewer of Netscape(tm)	   */
/*		browsers disk cache					   */
/*		Copyright (c) 1999,2000 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		modified 2005 ... 2012 by Harald Foerster		   */
/*		(harald_foerster@users.sourceforge.net)			   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#ifdef HAVE_GETOPT_LONG
#include <getopt.h>
#else
#include "getopt.h"
#endif

#ifdef HAVE_GLOB
#include <glob.h>
#endif

#ifdef BUILD_MINGW
#include <fcntl.h>
#include <windows.h>
#endif

#include "apassign.h"
#include "charconv.h"
#include "gui-gtk.h"
#include "file.h"
#include "gaccel.h"
#include "gprop.h"
#include "viewer.h"

/*
#include <gtk/gtk.h>
#include "nls.h"
#include "nscache.h"
*/


#define kPrintVersion	(1 << 0)
#define kPrintHelp	(1 << 1)

static const char *mozcacherc =
#ifdef BUILD_MINGW
			"mozcache.rc";
#else
			".mozcacherc";
#endif

static const char dir_separator[2] = { G_DIR_SEPARATOR, '\0' };

enum
{
	NSC_FIRST,
	NSC_TYPE = NSC_FIRST,
	NSC_INDEXDB,
	NSC_VIEWER,
	NSC_MENUACCEL,
	NSC_PROP,
	NSC_LAST = NSC_PROP
};

static const char *cfgrc_name[] =
{
	"CacheType:",
	"IndexDB:",
	"Viewer:",
	"MenuAccel:",
	"Property:"
};

static const char* nscache_cache_type_str[] =
{
#ifdef CONFIG_BERKELEY_DB
	"ns",
	"ns6",
#endif
	"moz",

#ifdef HAVE_MSIE
	"ie",
#endif
	"auto",
	NULL
};

static const char illegal_cache_type_str[] = gettext_nop("Illegal cache type");

static const char* nscache_sort_type_str[] =
{
	"url",
	"file",
	"size",
	"access",
	"mime",
	NULL
};

static int get_str_index(const char **tab, const char *pattern)
{
	int		idx;
	const char	*str;

	idx = 0;

	while ((str = *(tab++)) != NULL)
	{
		if (strcmp(str, pattern) == 0)
		{
			return idx;
		}

		idx++;
	}

	return -1; /* no match */

} /* static int get_str_index(const char**, const char*) */

static void fprn_stderr(const char *format, const char *mesg, const char *detail)
{

#ifdef BUILD_MINGW
	if ( ! main_open_console(TRUE))
	{
		return;
	}
#endif
	if (format == NULL)
	{
		format = "%s : %s\n";
	}

	fprintf(stderr, format, gettext(mesg), detail);

} /* static void fprn_stderr(const char*, const char*, const char*) */

#ifdef HAVE_GCONFIGDIR

static FILE *main_open_rc(const char *modus)
{
	FILE		*fd;
	char		*rc, *rc_config;
	const char	*home, *config;
#ifdef BUILD_MINGW
	const char	*config_utf8;
#endif
	config = g_get_user_config_dir();

	if (config != NULL)
	{
#ifdef BUILD_MINGW
		config_utf8 = config;
		config = CHARCONV_UTF8_TO_FILENAME(config_utf8);
		if (config == NULL)
		{
			config = config_utf8;
		}
#endif
		rc_config = g_strconcat(config,
#ifdef BUILD_MINGW
					dir_separator, mozcacherc, NULL);
#else
					dir_separator, &mozcacherc[1], NULL);
#endif
		rc = rc_config;

		if (access(rc_config, F_OK) == 0)
		{
#ifdef BUILD_MINGW
			CHARCONV_UTF8_FREE((char *) config, config_utf8);
#endif
			goto _open_file;
		}

		if (access(config, R_OK | W_OK) == -1)
		{
#ifdef BUILD_MINGW
			CHARCONV_UTF8_FREE((char *) config, config_utf8);
#endif
			g_free(rc_config);
			config = NULL;
		}
	}

	home = g_get_home_dir();

	if (home == NULL)
	{
		if (config != NULL)
		{
			goto _open_file;
		}

		return (FILE *) home;
	}

	if (config == NULL || strcmp(home, config) != 0)
	{
		char *rc_home;

#ifdef BUILD_MINGW
		const char *home_utf8 = home;
		home = CHARCONV_UTF8_TO_FILENAME(home_utf8);
		if (home == NULL)
		{
			home = home_utf8;
		}
#endif
		rc_home = g_strconcat(home, dir_separator, mozcacherc, NULL);

#ifdef BUILD_MINGW
		CHARCONV_UTF8_FREE((char *) home, home_utf8);
#endif
		if (config == NULL)
		{
			rc = rc_home;
		}

		else
		{
#ifdef BUILD_MINGW
			CHARCONV_UTF8_FREE((char *) config, config_utf8);
#endif
			if (access(rc_home, F_OK) == 0)
			{
#ifdef HAVE_LINK
				if (link(rc_home, rc_config) == -1)
#endif
				{
					static const char cpy_str[] =
#ifdef BUILD_MINGW
							"copy \"";
#else
							"cp -f \"";
#endif
					char *cpy = g_strconcat(cpy_str,
							rc_home, "\" \"",
							rc_config, "\"", NULL);
					system(cpy);
					g_free(cpy);
				}

				unlink(rc_home);
				g_free(rc_home);
			}
		}

	} /* if (config == NULL || strcmp(home, config) != 0 */

_open_file:
	fd = fopen(rc, modus);
	g_free(rc);

	return fd;

} /* static FILE *main_open_rc(const char*) */

#else  /* ! HAVE_GCONFIGDIR */

static FILE *main_open_rc(const char *modus)
{
	FILE		*fd;
	char		*rc;
	const char	*home;
#ifdef BUILD_MINGW
	const char	*home_utf8;
#endif
	home = g_get_home_dir();

	if (home == NULL)
	{
		return (FILE *) home;
	}

#ifdef BUILD_MINGW
	home_utf8	= home;
	home		= CHARCONV_UTF8_TO_FILENAME(home_utf8);
	if (home == NULL)
	{
		home = home_utf8;
	}
#endif
	rc = g_strconcat(home, dir_separator, mozcacherc, NULL);

#ifdef BUILD_MINGW
	CHARCONV_UTF8_FREE((char *) home, home_utf8);
#endif
	fd = fopen(rc, modus);
	g_free(rc);

	return fd;

} /* static FILE *main_open_rc(const char*) */

#endif  /* #ifdef HAVE_GCONFIGDIR .. #else */

char *main_save_rc(void)
{
	FILE *fd = main_open_rc("w");

	if (fd == NULL)
	{
		return gettext_nop("Unable to open the mozcacherc file");
	}

	fprintf(fd, "# %s\n# %s\n\n",
			gettext("This mozcacherc file was generated by MozCache"),
					gettext("You may edit it if you're careful!"));
	if (nscache_db.name)
	{
		const char *path;

		if (nscache_db.type < NSCACHE_RT_UNKNOWN)
		{
			fprintf(fd, "%s %s\n", cfgrc_name[NSC_TYPE],
					nscache_cache_type_str[nscache_db.type]);
		}

		path = nscache_db.path.str;

		fprintf(fd, "%s %s%s\n\n", cfgrc_name[NSC_INDEXDB],
				(path == NULL ? "" : path), nscache_db.name);
	}

	apassign_save_data(fd, cfgrc_name[NSC_VIEWER]);
	gaccel_save_data(fd, cfgrc_name[NSC_MENUACCEL]);
	gprop_save_data(fd, cfgrc_name[NSC_PROP]);

	fclose(fd);

	return NULL;

} /* char *main_save_rc(void) */

static void main_load_rc(int errmsg, char **dbname)
{
	FILE *fd;
	char *line, pom[2048];

	fd = main_open_rc("r");

	if (fd == NULL)
	{
		return;
	}

	while ((line = fgets(pom, sizeof(pom), fd)) != NULL)
	{
		static const char parse_msg[] = gettext_nop("Unable to parse");
		static const char parse_fmt[] = "%s : \"%s\"\n";
		static const char white_spaces[] = " \t";

		int i;

		line += strspn(line, white_spaces);

		if (*line == '#' || (i = strcspn(line, "\n")) == 0)
		{
			continue;
		}

		line[i]	= '\0';
		i	= NSC_LAST;

		do
		{
			int		len;
			const char	*str;

			str = cfgrc_name[i];
			len = strlen(str);

			if (GuiStrNCaseCmp(line, str, len) == 0)
			{
				line += len;
				line += strspn(line, white_spaces);
				break;
			}
		}
		while (--i >= NSC_FIRST);

		switch (i)
		{
			case NSC_TYPE:
			{
				int type = get_str_index(nscache_cache_type_str, line);

				if (type < 0)
				{
					if (errmsg)
					{
						fprn_stderr(NULL, illegal_cache_type_str, line);
					}

					break;
				}

				nscache_db.type = type;
				break;
			}

			case NSC_INDEXDB:
			{
				if (*line != '\0')
				{
					*dbname = g_strdup(line);
				}

				break;
			}

			case NSC_VIEWER:
			{
				apassign_parse_str(line);
				break;
			}

			case NSC_MENUACCEL:
			{
				if (gaccel_parse_str(line) == FALSE && errmsg)
				{
					fprn_stderr(parse_fmt, parse_msg, line);
				}

				break;
			}

			case NSC_PROP:
			{
				if (gprop_parse_str(line) == FALSE && errmsg)
				{
					fprn_stderr(parse_fmt, parse_msg, line);
				}

				break;
			}

			default:
			{
				if (errmsg)
				{
					fprn_stderr(parse_fmt, parse_msg, line);
				}

				break;
			}

		} /* switch (i) */

	} /* while ((line = fgets(pom, sizeof(pom), fd)) != NULL) */

	fclose(fd);

} /* static void main_load_rc(int, char**) */

static void print_version_help(int what, const char *prgname)
{

#ifdef CONFIG_BERKELEY_DB
#define BERKELEY_DB_STR	"ns|ns6|"
#else
#define BERKELEY_DB_STR
#endif

#ifdef HAVE_MSIE
#define MSIE_STR	"ie|"
#else
#define MSIE_STR
#endif

#ifdef BUILD_MINGW
	if ( ! main_open_console(TRUE))
	{
		return;
	}
#endif
	if (what & kPrintVersion)
	{
		printf("%s-%s\n", nscache_package_str, nscache_version_str);
	}

	if ( ! (what & kPrintHelp))
	{
		return;
	}

	printf("%s %s %s %s\n", gettext("Usage:"), prgname,
				gettext("[OPTION]"), gettext("[FILENAME]"));

	printf(
		"  -t,  --db-type=[" BERKELEY_DB_STR "moz|" MSIE_STR "auto]\n"
		"                           %s\n"
#ifdef CONFIG_BERKELEY_DB
		"                           ns   = %s\n"
		"                           ns6  = %s\n"
		"                                 %s\n"
#endif
		"                           moz  = %s\n"
		"                                 %s\n"
#ifdef HAVE_MSIE
		"                           ie   = %s\n"
#endif
		"                           auto = %s\n"
		"  -f,  --db-file=%s %s\n"
		"  -d,  --temp-dir=%s %s\n"
		"  -s,  --summary           %s\n"
		"  -c,  --content-types     %s\n"
		"  -l,  --list              %s\n"
		"  -m,  --multi-line        %s\n"
		"  -S,  --sort=[url|file|size|access|mime]\n"
		"                           %s\n"
		"  -R,  --sort-reverse      %s\n"
#ifndef CACHE_READONLY
		"  -r,  --readonly          %s\n"
#endif
#ifdef HAVE_GETTEXT
		"       --locale=[de_DE|es_ES|fr_FR|it_IT|ja_JP|pt_BR|sk_SK|tr_TR]\n"
		"       --nls-dir=%s %s\n"
#endif
		"  -v,  --version           %s\n"
		"  -h,  --help              %s\n",
		gettext("set cache type"),
#ifdef CONFIG_BERKELEY_DB
			gettext(NSC_GET_BROWSER_NAME(NSCACHE_RT_NETSCAPE_4)),
			gettext(NSC_GET_BROWSER_NAME(NSCACHE_RT_NETSCAPE_6)),
			gettext("(also Mozilla < 0.9)"),
#endif
			gettext(NSC_GET_BROWSER_NAME(NSCACHE_RT_MOZILLA)),
			gettext("(also Firefox, SeaMonkey and Netscape)"),
#ifdef HAVE_MSIE
			gettext(NSC_GET_BROWSER_NAME(NSCACHE_RT_MSIE)),
#endif
			gettext("unknown (default)"),
		gettext("FILENAME "),
			gettext("set cache file to FILENAME"),
		gettext("DIR     "),
			gettext("store temporary files in DIR (default: /tmp)"),
		gettext("print number of entries and total size"),
		gettext("print info about cache content types"),
		gettext("list cache directory contents to stdout"),
		gettext("print multi line info about each cache entry"),
		gettext("print cache entries sorted by type (default: url)"),
		gettext("print cache entries in reverse order"),
#ifndef CACHE_READONLY
		gettext("activate readonly mode"),
#endif
#ifdef HAVE_GETTEXT
		gettext("DIR      "),
			gettext("set the nlsdir (eg. /usr/share/locale)"),
#endif
		gettext("display version information and exit"),
		gettext("display this help and exit")

	); /* printf() */

	return;

} /* static void print_version_help(int, const char*) */

#if defined(DIRdata) || defined(HAVE_GLOB)

static char *get_mozilla_cache_dir(const char *path, const char *dir)
{
	char *sptr, *retv;

#ifdef DIRdata

	struct stat statbuf;

	sptr = g_strconcat(path, dir_separator, dir, NULL);
	retv = NULL;

	if (stat(sptr, &statbuf) == 0 && S_ISDIR(statbuf.st_mode))
	{
#ifdef HAVE_OPENDIR
		struct dirent *d;
#endif
		DIRdata *next = DIRopen(sptr);

		if (next)
		{
			const char *name;

			while ((name = DIRread(next, d)) != NULL)
			{
				if (name[0] == '.')
				{
					/* skip hidden directories */

					if (name[1] == '\0')
					{
						continue; /* => "." */
					}

					if (name[1] == '.' && name[2] == '\0')
					{
						continue; /* => ".." */
					}
				}

				if (strcmp(name, "Cache") == 0)
				{
					retv = g_strconcat(sptr,
						dir_separator, name, NULL);
					break;
				}

				if ((retv = get_mozilla_cache_dir(sptr, name)) != NULL)
				{
					break;
				}
			}

			DIRclose(next);
		}

	} /* if (stat(sptr, &statbuf) == 0 && S_ISDIR(statbuf.st_mode)) */

#elif defined(HAVE_GLOB)

	glob_t globdata;

	sptr = g_strconcat(path, dir, NULL);
	retv = NULL;
	
	if (glob(sptr, (GLOB_ERR | GLOB_MARK | GLOB_NOSORT),
						NULL, &globdata) == 0)
	{
		int i = globdata.gl_pathc;

		while (--i >= 0)
		{
			size_t		len;
			const char	*str;

			str = globdata.gl_pathv[i];
			len = strlen(str);

			if (str[len - 1] == G_DIR_SEPARATOR)
			{
#ifdef BUILD_MINGW
				static const char cache_str[] = "\\Cache\\";
#else
				static const char cache_str[] = "/Cache/";
#endif
				if (strcmp(cache_str, &str[len - (sizeof(cache_str) - 1)]) == 0)
				{
					retv = g_strdup(str);
					break;
				}

				if ((retv = get_mozilla_cache_dir(str, "*")) != NULL)
				{
					break;
				}
			}
		}

	} /* if (glob(path, (GLOB_ERR | GLOB_MARK | GLOB_NOSORT),
							NULL, &globdata) == 0) */
	globfree(&globdata);

#endif /* #ifdef HAVE_GDIROPEN .. #elif HAVE_GLOB */

	g_free(sptr);

	return retv;

} /* static char *get_mozilla_cache_dir(const char*, const char*) */

#endif /* #if defined(HAVE_GDIROPEN) || defined(HAVE_GLOB) */

#ifdef BUILD_MINGW

static int win_console = 0x00;

static int main_connect_console(void)
{
	int	newline;
	DWORD	which;

	newline	= 0;
	which	= STD_OUTPUT_HANDLE;

	for (;;)
	{
		int		fnr;
		HANDLE		stdHandle;
		FILE		*stdFp;

		stdFp = (which == STD_OUTPUT_HANDLE) ? stdout : stderr;

		if ((fnr = fileno(stdFp)) == -1 ||
			(stdHandle = (HANDLE) _get_osfhandle(fnr)) ==
							INVALID_HANDLE_VALUE ||
				GetFileType(stdHandle) == FILE_TYPE_UNKNOWN)
		{
			int	osfHandle;
			FILE	*fp;

			stdHandle = GetStdHandle(which);

			if (stdHandle == INVALID_HANDLE_VALUE)
			{
				break;
			}

			osfHandle = _open_osfhandle((long) stdHandle, _O_TEXT);

			if (osfHandle == -1)
			{
				break;
			}

			fp = _fdopen(osfHandle, "w" );

			if (fp == NULL)
			{
				break;
			}

			*stdFp = *fp;

			if (newline == FALSE)
			{
				newline = TRUE;
				fputs("\n", stdFp);
			}
		}

		if (which != STD_OUTPUT_HANDLE)
		{
			return TRUE;
		}

		which = STD_ERROR_HANDLE;

	} /* for (;;) */

	return FALSE;

} /* static int main_connect_console(void) */

static int main_attach_console(void)
{
	if (AttachConsole(ATTACH_PARENT_PROCESS))
	{
		win_console = 0x10;

		if (main_connect_console())
		{
			win_console = 0x11;
			return 0x01;
		}
	}

	return 0x00;

} /* static int main_attach_console(void) */

int main_open_console(int visible)
{
	if ( ! (win_console & 0x10) && ! main_attach_console())
	{
		win_console = 0x10;

		if ( ! AllocConsole())
		{
			return 0x00;
		}

		if (visible == FALSE)
		{
			char buf[1024];

			if (GetConsoleTitle(buf, sizeof(buf)) >= 0)
			{
				/*
					Give this a chance - it may
					fail the first time through
				*/

				int i = 10;

				do
				{
					HWND hwnd = FindWindowEx(NULL, NULL,
							NULL, (LPCTSTR) buf);

					if (hwnd != NULL)
					{
						ShowWindow(hwnd, SW_HIDE);
						break;
					}
				}
				while (--i > 0);
			}

			return 0x01;

		} /* if (visible == FALSE) */

		if (main_connect_console())
		{
			win_console = 0x11;
		}

	} /*  if ( ! (win_console & 0x10) && ! main_attach_console()) */

	return win_console & 0x01;

} /* int main_open_console(int) */

#endif /* BUILD_MINGW */

static int main_cmd_setup(int argc, char **argv, char **dbname)
{
	const struct option options[] =
	{
		{"db-type", 1, NULL, 't'},
		{"db-file", 1, NULL, 'f'},
		{"temp-dir", 1, NULL, 'd'},
		{"summary", 0, NULL, 's'},
		{"content-types", 0, NULL, 'c'},
		{"list", 0, NULL, 'l'},
		{"multi-line", 0, NULL, 'm'},
		{"sort", 1, NULL, 'S'},
		{"sort-reverse", 0, NULL, 'R'},
#ifndef CACHE_READONLY
		{"readonly", 0, NULL, 'r'},
#endif
#ifdef HAVE_GETTEXT
		{"locale", 1, NULL, 'g'},
		{"nls-dir", 1, NULL, 'g'},
#endif
		{"version", 0, NULL, 'v'},
		{"help", 0, NULL, 'h'},
		/* !!! GTK&GDK OPTIONS !!! */
		{"gtk-module", 1, NULL, 'g'},
		{"g-fatal-warnings", 0, NULL, 'g'},
		{"gtk-debug", 2, NULL, 'g'},
		{"gtk-no-debug", 2, NULL, 'g'},
		{"gdk-debug", 2, NULL, 'g'},
		{"gdk-no-debug", 2, NULL, 'g'},
		{"display", 1, NULL, 'g'},
		{"sync", 0, NULL, 'g'},
		{"no-xshm", 0, NULL, 'g'},
		{"name", 1, NULL, 'g'},
		{"class", 1, NULL, 'g'},
		{"gxid_host", 1, NULL, 'g'},
		{"gxid_port", 1, NULL, 'g'},
		{"xim-preedit", 1, NULL, 'g'},
		{"xim-status", 1, NULL, 'g'},
		{NULL, 0, NULL, 0},
	};

	/* viewer.c */
	extern char *viewer_temp_dir;

	int	c, index, print_list, print_info, sort_reverse;
	char	*db_name;
	char	*db_type;
	char	*sort_type;
	char	*temp_dir;

#ifdef BUILD_MINGW
	char	*temp_utf8;

	if (argc <= 1 || ! main_attach_console())
	{
		/* disable error messages - no 'stderr' */
		opterr = 0;
	}
#endif

	print_list = kNoStdOut;
	print_info = 0;
	sort_reverse = 0;

	db_name	   = NULL;
	db_type    = NULL;
	sort_type  = NULL;
	temp_dir   = NULL;

	*dbname		= NULL;
	nscache_db.type	= NSCACHE_RT_UNKNOWN;

#ifndef CACHE_READONLY
#define OPTION_RDONLY	"r"
#else
#define OPTION_RDONLY
#endif
	while ((c = getopt_long(argc, argv, OPTION_RDONLY "sclmvhRd:f:t:S:",
							options, &index)) != - 1)
	{
		switch (c)
		{
			case 't':
				db_type = optarg;
				break;
			case 'f':
				db_name = optarg;
				break;
			case 'd':
				temp_dir = optarg;
				break;
			case 's':
				print_list = kOnlyStdOutSummary;
				break;
			case 'c':
				print_list = kOnlyStdOutConentTypes;
				break;
			case 'l':
				print_list = kOnlyStdOutList;
				break;
			case 'm':
				print_list = kOnlyStdOutMultiLine;
				break;
			case 'S':
				sort_type = optarg;
				break;
			case 'R':
				sort_reverse = STYPE_REVERS_MASK;
				break;
			case 'v':
				print_info |= kPrintVersion;
				break;
			case 'h':
				print_info |= kPrintHelp;
				break;
#ifndef CACHE_READONLY
			case 'r':
				nscache_db.readonly = TRUE;
				break;
#endif
			case 'g':
				break;

			default:
				return EINVAL;

		} /* switch (c) */

	} /* while ((c = getopt_long(argc, argv, ... )) != - 1) */

	if (optind < argc)
	{
		if (db_name != NULL || argc > optind + 1)
		{
			fprn_stderr(NULL, gettext_nop("Illegal argument"), argv[optind]);
			return EINVAL;
		}

		db_name = argv[optind];
	}

	if (print_info)
	{
		print_version_help(print_info, argv[0]);
		exit(0);
	}

	main_load_rc(print_list == kNoStdOut, dbname);

	if (db_name)
	{
		g_free(*dbname);

		*dbname		= g_strdup(db_name);
		nscache_db.type	= NSCACHE_RT_UNKNOWN;
	}

	if (db_type)
	{
		int type = get_str_index(nscache_cache_type_str, db_type);

		if (type < 0)
		{
			fprn_stderr(NULL, illegal_cache_type_str, db_type);
			return EINVAL;
		}

		nscache_db.type = type;
	}

	if (print_list)
	{
		int error, stype;

#ifdef BUILD_MINGW
		if ( ! main_open_console(TRUE))
		{
			exit(0);
		}
#endif

		if (*dbname == NULL
#ifdef HAVE_MSIE
			&& (nscache_db.type != NSCACHE_RT_MSIE)
#endif
			)
		{
			fprn_stderr("%s%s", gettext_nop("Cache file not specified!"), "\n");
			return EINVAL;
		}

		stype = STYPE_URL;

		if (sort_type && (print_list == kOnlyStdOutList ||
					print_list == kOnlyStdOutMultiLine))
		{
			stype = get_str_index(nscache_sort_type_str, sort_type);

			if (stype < 0)
			{
				fprn_stderr(NULL, gettext_nop("Illegal sort type"), sort_type);
				return EINVAL;
			}
		}

		error = nscache_dump(*dbname, print_list, stype | sort_reverse);
		exit(error);

	} /* if (print_list) */

	if (temp_dir)
	{
		struct stat	tmpstat;
		const char	*dir = NULL;

		if (temp_dir[0] == '\0')
		{
			dir = g_get_current_dir();
		}

		else
		{
			if (temp_dir[0] != '/'
#ifdef BUILD_MINGW
			|| (temp_dir[0] != '\\' && temp_dir[1] != ':')
#endif
				) /* if (temp_dir[0] != '/' */
			{

				if (temp_dir[0] == '~' &&
					(temp_dir[1] == '\0' ||
							temp_dir[1] == '/'))
				{
					temp_dir++;
					dir = g_get_home_dir();
				}

				else
				{
					dir = g_get_current_dir();
				}
			}
		}

		if (dir != NULL)
		{

#ifdef BUILD_MINGW
			temp_utf8 = (char *) dir;
			dir = CHARCONV_UTF8_TO_FILENAME(temp_utf8);
			if (dir == NULL)
			{
				dir = temp_utf8;
			}
#endif
			temp_dir = g_strconcat(dir, dir_separator,
							temp_dir, NULL);
#ifdef BUILD_MINGW
			CHARCONV_UTF8_FREE((char *) dir, temp_utf8);
#endif
		} /* if (dir != NULL) */

		if ( stat(temp_dir, &tmpstat) ||
			! (S_ISDIR(tmpstat.st_mode)) ||
				access(temp_dir, R_OK | W_OK))
		{
			g_free(temp_dir);
			temp_dir = NULL;
		}

	} /* if (temp_dir) */

	if (temp_dir == NULL)
	{
		temp_dir = strdup(g_get_tmp_dir());

#ifdef BUILD_MINGW
		temp_utf8 = temp_dir;
		temp_dir  = CHARCONV_UTF8_TO_FILENAME(temp_utf8);
		if (temp_dir == NULL)
		{
			temp_dir = temp_utf8;
		}
		else if (temp_utf8 != temp_dir)
		{
			g_free(temp_utf8); /* strdup'd */
		}
#endif
	}

	if (temp_dir)
	{
		int len = strlen(temp_dir) - 1;

#ifdef BUILD_MINGW
		file_path_dos_to_unix(temp_dir);
#endif
		while (len >= 0 && temp_dir[len] == '/')
		{
			temp_dir[len] = '\0';
			len--;
		}

		viewer_temp_dir = temp_dir;

	} /* if (temp_dir) */


	if (*dbname == NULL)
	{
		char *path, **ptr, *config[3];

		ptr = config;

#ifdef HAVE_GCONFIGDIR
		if ((*ptr = g_strdup(g_get_user_config_dir())) != NULL)
		{
			ptr++;
		}
#endif
		if ((*ptr = g_strdup(g_get_home_dir())) != NULL)
		{
			ptr++;
		}

		*ptr = NULL;


#if defined(DIRdata) || defined(HAVE_GLOB)

		ptr = config;

		while ((path = *(ptr++)) != NULL)
		{
			static const char *moz_path =

#ifdef DIRdata
#ifdef BUILD_MINGW
					"Mozilla";
#else
					".mozilla";
#endif
#else /* HAVE_GLOB */
#ifdef BUILD_MINGW
					"\\Mozilla\\*";
#else
					"/.mozilla/*";
#endif
#endif

#ifdef BUILD_MINGW
			char *path_utf8 = path;
			path = CHARCONV_UTF8_TO_FILENAME(path_utf8);
			if (path == NULL)
			{
				path = path_utf8;
			}
#endif
			*dbname = get_mozilla_cache_dir(path, moz_path);
#ifdef BUILD_MINGW
			CHARCONV_UTF8_FREE(path, path_utf8);
#endif
			if (*dbname != NULL)
			{
				break;
			}

		} /* while ((path = *(ptr++)) != NULL) */

#endif /* defined(DIRdata) || defined(HAVE_GLOB) */


		if (*dbname == NULL)
		{

#ifdef CONFIG_BERKELEY_DB
			ptr = config;

			while ((path = *(ptr++)) != NULL)
			{
				*dbname = g_strconcat(path,
						"/.netscape/cache", NULL);

				if (access(*dbname, F_OK) != 0)
				{
					break;
				}

				g_free(*dbname);
			}

#endif /* CONFIG_BERKELEY_DB */

			if (*dbname == NULL && config[0] != NULL)
			{
				*dbname = g_strconcat(config[0],
							"/.mozilla", NULL);
			}

		} /* if (*dbname == NULL) */

		ptr = config;

		while ((path = *(ptr++)) != NULL)
		{
			g_free(path);
		}

	} /* if (*dbname == NULL) */

	return 0;

} /* static int main_cmd_setup(int, char**, char**) */

static void main_sigterm(int signum)
{
	int save_flag;

	/* store settings */
	if ( ! gprop_get_bool(kPropBoolOptSaveConfig, &save_flag) || save_flag)
	{
		const char *errstr = main_save_rc();

		if (errstr && signum != SIGTERM)
		{
			gui_strerror(errstr);
		}
	}

	/* Remove all temporary files */
	viewer_atexit();

	if (signum == SIGTERM)
	{
		_exit(0);
	}
}

static void main_atexit(void)
{
	/* user lost interrest */
	main_sigterm(SIGQUIT);
}

static void main_signal_init(void)
{
#ifdef HAVE_SIGACTION

	struct sigaction sigact;

	parent_pid = getpid();
	parent_uid = getuid();

	/* block other signals, while executing 'sa_handler' */
	sigfillset(&sigact.sa_mask);
	sigact.sa_flags = 0;

	/* ignore broken pipes */
	sigact.sa_handler = SIG_IGN;
	sigaction(SIGPIPE, &sigact, NULL);

	/* MozCache was terminated */
	sigact.sa_handler = main_sigterm;
	sigaction(SIGTERM, &sigact, NULL);

	/* child was terminated */
	sigact.sa_handler = viewer_sigchld;
	sigact.sa_flags = SA_NOCLDSTOP;
	sigaction(SIGCHLD, &sigact, NULL);
#else
	signal(SIGTERM, main_sigterm);
#endif

} /* static void main_signal_init(void) */

#ifdef HAVE_GSETENV
#define SETENV(name, value) g_setenv((name), (value), TRUE)
#else
#define SETENV(name, value) putenv(g_strconcat((name), "=", (value), NULL))
#endif

static void main_init_locale(int argc, const char **argv)
{
	const char *arg, *locale, *nlsdir;

	locale = nlsdir = NULL;

	while ((arg = *(argv++)) != NULL)
	{
		if (strncmp(arg, "--locale", 8) == 0)
		{
			if (arg[8] == '=')
			{
				/* --locale=LOCALE */
				locale = &arg[9];
			}

			else
			{
				/* --locale LOCALE */
				locale = *(argv++);
			}
		}

		else if (strncmp(arg, "--nls-dir", 9) == 0)
		{
			if (arg[9] == '=')
			{
				/* --nls-dir=DIR */
				nlsdir = &arg[10];
			}

			else
			{
				/* --nls-dir DIR */
				nlsdir = *(argv++);
			}
		}
	}

	if (locale)
	{
		const char *s;

		SETENV("LANG", locale);

		s = strchr(locale, '.');

		if (s)
		{
			s = g_strndup(locale, s - locale);
		}

		else
		{
			s = locale;
		}

		SETENV("LANGUAGE", s);

		if (strchr(s, '_'))
		{
			SETENV("LC_ALL", s);
		}

		if (s != locale)
		{
			g_free((char *) s);
		}
	}

	NLS_INIT_LOCALE(nlsdir);

} /* static void main_init_locale(int, const char**) */

int main(int argc, char **argv)
{
	int	error;
	char	*dbname;

	main_init_locale(argc, (const char **) argv);

	error = main_cmd_setup(argc, argv, &dbname);

	if (error)
	{
		return error;
	}

#ifdef HAVE_LINK
	gprop_get_bool(kPropBoolSaveHardlink, &hardlink);
#endif
	gprop_get_bool(kPropBoolSaveSingleFolder, &single_folder);
	gprop_get_int(kPropIntSaveUrlToLocal, &url_to_local_selector);

	apassign_set_default_viewers();

	gtk_init(&argc, &argv);
	charconv_init();
	gui_init(dbname);

	main_signal_init();
	atexit(main_atexit);

	gtk_main();

	return 0;

} /* int main(int, char**) */

/* EOF */
